#include <fstream>
#include <iostream>

#include "EtsiTs103097Module.hh"

#include "certificates_loader.hh"

#include "etsi_ts103097_certificate_codec.hh"
#include "etsi_ts103097_tobesigned_certificate_codec.hh"

#include "security_services_its.hh"

#include "sha256.hh"
#include "sha384.hh"

#include "converter.hh"

#include "loggers.hh"

bool security_cache_comp::operator()(const OCTETSTRING &p_lhs, const OCTETSTRING &p_rhs) const {
  // loggers::get_instance().log_msg(">>> security_cache_comp::operator(): p_lhs= ", p_lhs);
  // loggers::get_instance().log_msg(">>> security_cache_comp::operator(): p_rhs= ", p_rhs);
  unsigned char *first1 = (unsigned char *)static_cast<const unsigned char *>(p_lhs);
  unsigned char *first2 = (unsigned char *)static_cast<const unsigned char *>(p_rhs);
  unsigned char *last1  = p_lhs.lengthof() + (unsigned char *)static_cast<const unsigned char *>(p_lhs);
  unsigned char *last2  = p_rhs.lengthof() + (unsigned char *)static_cast<const unsigned char *>(p_rhs);

  while (first1 != last1) {
    if ((first2 == last2) || (*first2 < *first1)) {
      return false;
    } else if (*first1 < *first2) {
      return true;
    }
    ++first1;
    ++first2;
  } // End of 'while' statement
  return (first2 != last2);
}

certificates_loader *certificates_loader::instance = nullptr;

certificates_loader::certificates_loader()
  : _certificateExt{".oer"}, _privateKeyExt{".vkey"}, _privateEncKeyExt{".ekey"},
    _full_path(), _is_cache_initialized{false}, _directory_filter{".svn", "._.DS_Store", ".DS_Store"}, _hashed_id8s(), _certificates_idx() {
  // loggers::get_instance().log(">>> certificates_loader::certificates_loader");
} // End of ctor

int certificates_loader::build_path(const std::string &p_root_directory) {
  // loggers::get_instance().log(">>> certificates_loader::build_path: '%s'", p_root_directory.c_str());

  // Build full path
  if (!p_root_directory.empty()) {
    _full_path = p_root_directory;
    if (!std::experimental::filesystem::exists(_full_path) ||
        !std::experimental::filesystem::is_directory(_full_path)) { // FIXME Coredump when app hasn't the rights to create the directory!!!!
      // Create directory
      if (!std::experimental::filesystem::create_directory(_full_path)) {
        _full_path = "./";
      } else { // Set rights for all users
        std::experimental::filesystem::permissions(_full_path,
                                                   std::experimental::filesystem::perms::add_perms | std::experimental::filesystem::perms::owner_all |
                                                     std::experimental::filesystem::perms::group_all | std::experimental::filesystem::perms::others_all);
      }
    }
  } else {
    _full_path = "./";
  }
  std::experimental::filesystem::canonical(_full_path);
  // loggers::get_instance().log("certificates_loader::build_path: full path: %s", _full_path.string().c_str());
  if (!std::experimental::filesystem::exists(_full_path)) {
    loggers::get_instance().warning("certificates_loader::build_path: Invalid path");
    _full_path.clear();
    return -1;
  }

  return 0;
} // End of method build_path

int certificates_loader::build_certificates_index() { // FIXME FSCOM Merge build_path and build_certificates_index methods in one
  loggers::get_instance().log(">>> certificates_loader::build_certificates_index");

  // Sanity check
  if (_is_cache_initialized) {
    return 0;
  }

  // Load index file
  std::experimental::filesystem::path p = _full_path.string() + "/index.lst";
  loggers::get_instance().log("certificates_loader::build_certificates_index: Caching certificate '%s'", p.string().c_str());
  std::ifstream is(p, ios::in);
  if (!is.is_open()) {
    loggers::get_instance().warning("certificates_loader::build_certificates_index: Failed to open index file");
    return -1;
  }
  std::string hashed_id8, certificate;
  while (is >> hashed_id8 >> certificate) { // Process lines
    //loggers::get_instance().log("certificates_loader::build_certificates_index: Processing  '%s - %s'", hashed_id8.c_str(), certificate.c_str());
    // Convert HashedId8 into oectet string_to_into
    OCTETSTRING h = str2oct(CHARSTRING(static_cast<const char*>(hashed_id8.c_str())));
    // Remove extension file
    p = certificate;
    certificate = p.stem();
    // Update index maps
    std::pair<std::map<OCTETSTRING, const std::string>::iterator, bool> r1 = _hashed_id8s.insert(std::pair<OCTETSTRING, const std::string>(h, certificate));
    if (r1.second == false) {
      loggers::get_instance().warning("certificates_loader::build_certificates_index (1): Failed to insert new record '%s'", certificate.c_str());
      continue;
    }
    std::pair<std::map<std::string, const OCTETSTRING>::iterator, bool> r2 = _certificates_idx.insert(std::pair<std::string, const OCTETSTRING>(certificate, h));
    if (r2.second == false) {
      loggers::get_instance().warning("certificates_loader::build_certificates_index (2): Failed to insert new record '%s'", certificate.c_str());
      continue;
    }
  } // End of 'while' statement
   is.close();

  _is_cache_initialized = true;

  return 0;
} // End of method build_certificates_index

std::map<OCTETSTRING, const std::string, security_cache_comp>& certificates_loader::get_hashed_id8s() {
  return _hashed_id8s;
}

std::map<std::string, const OCTETSTRING>& certificates_loader::get_certificates() {
  return _certificates_idx;
}

int certificates_loader::load_certificate(const OCTETSTRING& p_hashed_id8, const std::string p_certificate_name, std::map<std::string, std::unique_ptr<const security_db_record>> & p_certificates) {
  loggers::get_instance().log_msg(">>> certificates_loader::load_certificate: ", p_hashed_id8);
  loggers::get_instance().log(">>> certificates_loader::load_certificate: '%s'", p_certificate_name.c_str());

  std::experimental::filesystem::path p = _full_path.string() + "/" + p_certificate_name + _certificateExt;
  const std::string& key = p.stem();
  loggers::get_instance().log("certificates_loader::load_certificate: Caching certificate '%s'", p.string().c_str());
  std::ifstream is(p.string().c_str(), ios::in | ios::binary);
  if (!is.is_open()) {
    loggers::get_instance().warning("certificates_loader::load_certificate: Failed to open certificate file");
    return -1;
  }
  OCTETSTRING certificate = int2oct(0, std::experimental::filesystem::file_size(p));
  is.read((char *)static_cast<const unsigned char *>(certificate), certificate.lengthof());
  is.close();
  loggers::get_instance().log_msg("certificates_loader::load_certificate: certificate: ", certificate);

  // Load private key file
  OCTETSTRING private_key;
  p = p.replace_extension(_privateKeyExt);
  // loggers::get_instance().log("certificates_loader::load_certificate: Caching private keys '%s'", p.string().c_str());
  is.open(p, ios::in | ios::binary);
  if (!is.is_open()) {
    loggers::get_instance().warning("certificates_loader::load_certificate: Failed to open Private key");
    private_key = OCTETSTRING(0, nullptr);
  } else {
    int size = std::experimental::filesystem::file_size(p);
    if ((size != 32) && (size != 48)) {
      loggers::get_instance().warning("certificates_loader::load_certificate: Private key size is incorrect for '%s'", key.c_str());
      return -1;
    }
    private_key = int2oct(0, size);
    is.read((char *)static_cast<const unsigned char *>(private_key), private_key.lengthof());
    is.close();
  }

  // Load private encryption key file if present
  OCTETSTRING private_enc_key;
  p = p.replace_extension(_privateEncKeyExt);
  // loggers::get_instance().log("certificates_loader::load_certificate: Caching private encryption key '%s'", p.string().c_str());
  is.open(p, ios::in | ios::binary);
  if (!is.is_open()) {
    loggers::get_instance().warning("certificates_loader::load_certificate: Failed to open Private encryption key file");
    private_enc_key = OCTETSTRING(0, nullptr);
  } else {
    int size = std::experimental::filesystem::file_size(p);
    if (size != 32) { // IEEE Std 1609.2 2017: NistP256 or BrainpoolP256r1 or Sm2P256
      loggers::get_instance().warning("certificates_loader::load_certificate: Private encryption key size is incorrect for '%s'", key.c_str());
      return -1;
    }
    private_enc_key = int2oct(0, size);
    is.read((char *)static_cast<const unsigned char *>(private_enc_key), private_enc_key.lengthof());
    is.close();
  }

  // Whole-certificate hash using SHA-256
  OCTETSTRING hash_sha_256;
  sha256      sha;
  sha.generate(certificate, hash_sha_256);
  loggers::get_instance().log_msg("certificates_loader::load_certificate: Whole-certificate SHA-256 hash for encryption: ", hash_sha_256);

  std::pair<std::map<std::string, std::unique_ptr<const security_db_record>>::iterator, bool> result;
  // Build DB record
  etsi_ts103097_certificate_codec codec;
  Ieee1609Dot2::CertificateBase   decoded_certificate;
  codec.decode(certificate, decoded_certificate);
  if (!decoded_certificate.is_bound()) {
    loggers::get_instance().warning("certificates_loader::load_certificate: Failed to decode certificate for '%s'", key.c_str());
    return -1;
  } else {
    // Extract common part of the certificate
    OCTETSTRING public_enc_key_x;
    OCTETSTRING public_enc_key_y;
    OCTETSTRING public_enc_comp_key;
    if (decoded_certificate.toBeSigned().encryptionKey().ispresent()) {
      Ieee1609Dot2BaseTypes::PublicEncryptionKey &v =
        static_cast<Ieee1609Dot2BaseTypes::PublicEncryptionKey &>(decoded_certificate.toBeSigned().encryptionKey());
      Ieee1609Dot2BaseTypes::BasePublicEncryptionKey &b = v.publicKey();
      if (b.ischosen(Ieee1609Dot2BaseTypes::BasePublicEncryptionKey::ALT_eciesNistP256)) {
        Ieee1609Dot2BaseTypes::EccP256CurvePoint &p = v.publicKey().eciesNistP256();
        security_services_its::fill_public_key_vectors(ec_elliptic_curves::nist_p_256, p, public_enc_comp_key, public_enc_key_x, public_enc_key_y);
      } else if (b.ischosen(Ieee1609Dot2BaseTypes::BasePublicEncryptionKey::ALT_eciesBrainpoolP256r1)) {
        Ieee1609Dot2BaseTypes::EccP256CurvePoint &p = v.publicKey().eciesBrainpoolP256r1();
        security_services_its::fill_public_key_vectors(ec_elliptic_curves::brainpool_p_256_r1, p, public_enc_comp_key, public_enc_key_x, public_enc_key_y);
      } else if (b.ischosen(Ieee1609Dot2BaseTypes::BasePublicEncryptionKey::ALT_ecencSm2)) {
        Ieee1609Dot2BaseTypes::EccP256CurvePoint &p = v.publicKey().ecencSm2();
        security_services_its::fill_public_key_vectors(ec_elliptic_curves::sm2_p_256, p, public_enc_comp_key, public_enc_key_x, public_enc_key_y);
      }
      // loggers::get_instance().log_msg("certificates_loader::load_certificate: public_enc_key_x: ", public_enc_key_x);
      // loggers::get_instance().log_msg("certificates_loader::load_certificate: public_enc_key_y: ", public_enc_key_y);
      // loggers::get_instance().log_msg("certificates_loader::load_certificate: public_enc_comp_key: ", public_enc_comp_key);
    } else {
      public_enc_key_x    = OCTETSTRING(0, nullptr);
      public_enc_key_y    = OCTETSTRING(0, nullptr);
      public_enc_comp_key = OCTETSTRING(0, nullptr);
    }

    OCTETSTRING issuer;                  // Certificate issuer
    if (decoded_certificate.issuer().ischosen(Ieee1609Dot2::IssuerIdentifier::ALT_sha256AndDigest)) {
      issuer = decoded_certificate.issuer().sha256AndDigest();
    } else if (decoded_certificate.issuer().ischosen(Ieee1609Dot2::IssuerIdentifier::ALT_sha384AndDigest)) {
      issuer = decoded_certificate.issuer().sha384AndDigest();
    } else if (decoded_certificate.issuer().ischosen(Ieee1609Dot2::IssuerIdentifier::ALT_self__)) {
      OCTETSTRING h;
      if (decoded_certificate.issuer().self__() == Ieee1609Dot2BaseTypes::HashAlgorithm::sha256) {
        sha256 sha;
        sha.generate(certificate, h);
        // loggers::get_instance().log_msg("certificates_loader::load_certificate: Whole-certificate SHA-256 hash: ", hash);
      } else {
        sha384 sha;
        sha.generate(certificate, h);
        // loggers::get_instance().log_msg("certificates_loader::load_certificate: Whole-certificate SHA-384 hash: ", hash);
      }
      issuer = OCTETSTRING(8, static_cast<const unsigned char *>(h) + h.lengthof() - 8);
    } else {
      issuer = int2oct(0, 8);
    }
    loggers::get_instance().log_msg("certificates_loader::load_certificate: issuer: ", issuer);

    if (decoded_certificate.toBeSigned().verifyKeyIndicator().ischosen(Ieee1609Dot2::VerificationKeyIndicator::ALT_verificationKey)) {
      // loggers::get_instance().log_msg("certificates_loader::load_certificate: Decoded certificate: ", decoded_certificate);
      ec_elliptic_curves algorithm; 
      OCTETSTRING public_key_x;
      OCTETSTRING public_key_y;
      OCTETSTRING public_comp_key; // public compressed key, 33 or 49 bytes length, byte #0 indicating compressed-y-0 (0x02) or compressed-y-1 (0x03)
      Ieee1609Dot2BaseTypes::PublicVerificationKey &b = decoded_certificate.toBeSigned().verifyKeyIndicator().verificationKey();
      if (b.ischosen(Ieee1609Dot2BaseTypes::PublicVerificationKey::ALT_ecdsaNistP256)) {
        algorithm = ec_elliptic_curves::nist_p_256;
        Ieee1609Dot2BaseTypes::EccP256CurvePoint &p = b.ecdsaNistP256();
        security_services_its::fill_public_key_vectors(ec_elliptic_curves::nist_p_256, p, public_comp_key, public_key_x, public_key_y);
      } else if (b.ischosen(Ieee1609Dot2BaseTypes::PublicVerificationKey::ALT_ecdsaNistP384)) {
        algorithm = ec_elliptic_curves::nist_p_384;
        Ieee1609Dot2BaseTypes::EccP384CurvePoint &p = b.ecdsaNistP384();
        security_services_its::fill_public_key_vectors(ec_elliptic_curves::nist_p_384, p, public_comp_key, public_key_x, public_key_y);
      } else if (b.ischosen(Ieee1609Dot2BaseTypes::PublicVerificationKey::ALT_ecdsaBrainpoolP256r1)) {
        algorithm = ec_elliptic_curves::brainpool_p_256_r1;
        Ieee1609Dot2BaseTypes::EccP256CurvePoint &p = b.ecdsaBrainpoolP256r1();
        security_services_its::fill_public_key_vectors(ec_elliptic_curves::brainpool_p_256_r1, p, public_comp_key, public_key_x, public_key_y);
      } else if (b.ischosen(Ieee1609Dot2BaseTypes::PublicVerificationKey::ALT_ecdsaBrainpoolP384r1)) {
        algorithm = ec_elliptic_curves::brainpool_p_384_r1;
        Ieee1609Dot2BaseTypes::EccP384CurvePoint &p = b.ecdsaBrainpoolP384r1();
        security_services_its::fill_public_key_vectors(ec_elliptic_curves::brainpool_p_384_r1, p, public_comp_key, public_key_x, public_key_y);
      } else { // ALT_ecsigSm2
        algorithm = ec_elliptic_curves::sm2_p_256;
        Ieee1609Dot2BaseTypes::EccP256CurvePoint &p = b.ecsigSm2();
        security_services_its::fill_public_key_vectors(ec_elliptic_curves::sm2_p_256, p, public_comp_key, public_key_x, public_key_y);
      }
      // loggers::get_instance().log("certificates_loader::load_certificate: algorithm: %d", algorithm);
      // loggers::get_instance().log_msg("certificates_loader::load_certificate: public_key_x: ", public_key_x);
      // loggers::get_instance().log_msg("certificates_loader::load_certificate: public_key_y: ", public_key_y);
      // loggers::get_instance().log_msg("certificates_loader::load_certificate: public_comp_key: ", public_comp_key);

      OCTETSTRING hash;                    // Whole-certificate hash
      OCTETSTRING hashed_id;               // Whole-certificate hashedid-8
      if (public_key_x.lengthof() == 32) { // See IEEE Std 1609.2a-2017 Clause 6.4.3 CertificateBase
        hash = hash_sha_256;
        loggers::get_instance().log_msg("certificates_loader::load_certificate: Whole-certificate SHA-256 hash: ", hash);
        hashed_id = OCTETSTRING(8, static_cast<const unsigned char *>(hash) + hash.lengthof() - 8);
      } else if (public_key_x.lengthof() == 48) {
        sha384 sha;
        sha.generate(certificate, hash);
        loggers::get_instance().log_msg("certificates_loader::load_certificate: Whole-certificate SHA-384 hash: ", hash);
        hashed_id = OCTETSTRING(8, static_cast<const unsigned char *>(hash) + hash.lengthof() - 8);
      } else {
        hash         = OCTETSTRING(0, nullptr);
        hash_sha_256 = OCTETSTRING(0, nullptr);
        hashed_id    = int2oct(0, 8);
      }
      loggers::get_instance().log_msg("certificates_loader::load_certificate: hash: ", hash);
      loggers::get_instance().log_msg("certificates_loader::load_certificate: hashed_id: ", hashed_id);

      // Create new record
      security_db_record *r = new security_db_record(
        algorithm,          // Signing algorithm
        key,                // Certificate name. Index key
        certificate,        // Certificate
        decoded_certificate,
        issuer,             // Hashed ID fo the issuer
        hash,               // Hash of the certificate
        hash_sha_256,       // Whole-certificate SHA256
        hashed_id,          // Hashed ID
        private_key,        // Private key
        public_key_x,       // public keys X-coordinate
        public_key_y,       // public keys Y-coordinate
        public_comp_key,    // public compressed key, 33 or 49 bytes length, byte #0 indicating compressed-y-0 (0x02) or compressed-y-1 (0x03)
        private_enc_key,    // Private encryption key
        public_enc_key_x,   // Public encryption key X-coordinate
        public_enc_key_y,   // Public encryption key Y-coordinate
        public_enc_comp_key // Public compressed encryption key
      );
      result = p_certificates.insert(std::pair<std::string, std::unique_ptr<const security_db_record>>(key, std::unique_ptr<const security_db_record>(r)));
      if (result.second == false) {
        loggers::get_instance().warning("certificates_loader::load_certificate: Failed to insert new record '%s'", key.c_str());
        delete r;
        return -1;
      }
    } else if (decoded_certificate.toBeSigned().verifyKeyIndicator().ischosen(Ieee1609Dot2::VerificationKeyIndicator::ALT_reconstructionValue)) {
      // Load certificate issuer
      std::map<OCTETSTRING, const std::string>::const_iterator it = _hashed_id8s.find(issuer);
      if (it == _hashed_id8s.cend()) {
        loggers::get_instance().warning("certificates_loader::load_certificate: Cannot find issuer of implicit certificate");
        return -1;
      }
      std::map<std::string, std::unique_ptr<const security_db_record>> m;
      if (load_certificate(issuer, it->second, m) == -1) {
        loggers::get_instance().warning("certificates_loader::load_certificate: Cannot load implicit certificate issuer");
        return -1;
      }
      // The signing algorithm is provided by the issuer certificate
      if (!m.cbegin()->second->decoded_certificate().toBeSigned().verifyKeyIndicator().ischosen(Ieee1609Dot2::VerificationKeyIndicator::ALT_verificationKey)) {
        loggers::get_instance().warning("certificates_loader::load_certificate: Inconsistent certificate issuer: Wrong VerificationKeyIndicator");
        return -1;      
      }
      loggers::get_instance().log_msg("certificates_loader::load_certificate: Implicit certificate issuer: ", m.cbegin()->second->decoded_certificate());
      // Retrieve the issuer signing key
      const Ieee1609Dot2BaseTypes::PublicVerificationKey &b = m.cbegin()->second->decoded_certificate().toBeSigned().verifyKeyIndicator().verificationKey();
      ec_elliptic_curves algorithm; 
      OCTETSTRING issuer_sign_key_x;
      OCTETSTRING issuer_sign_key_y;
      OCTETSTRING issuer_sign_key_comp_key; // public compressed key, 33 or 49 bytes length, byte #0 indicating compressed-y-0 (0x02) or compressed-y-1 (0x03)
      if (b.ischosen(Ieee1609Dot2BaseTypes::PublicVerificationKey::ALT_ecdsaNistP256)) {
        const Ieee1609Dot2BaseTypes::EccP256CurvePoint &p = b.ecdsaNistP256();
        security_services_its::fill_public_key_vectors(ec_elliptic_curves::nist_p_256, p, issuer_sign_key_comp_key, issuer_sign_key_x, issuer_sign_key_y);
        algorithm = ec_elliptic_curves::nist_p_256;
      } else if (b.ischosen(Ieee1609Dot2BaseTypes::PublicVerificationKey::ALT_ecdsaBrainpoolP256r1)) {
        const Ieee1609Dot2BaseTypes::EccP256CurvePoint &p = b.ecdsaBrainpoolP256r1();
        security_services_its::fill_public_key_vectors(ec_elliptic_curves::brainpool_p_256_r1, p, issuer_sign_key_comp_key, issuer_sign_key_x, issuer_sign_key_y);
        algorithm = ec_elliptic_curves::brainpool_p_256_r1;
      } else {
        loggers::get_instance().warning("certificates_loader::load_certificate: Incosistent certificate issuer: Wrong PublicVerificationKey");
        return -1;      
      }
      // Retrieve the reconstruction key key
      const Ieee1609Dot2BaseTypes::EccP256CurvePoint &p = decoded_certificate.toBeSigned().verifyKeyIndicator().reconstructionValue();
      OCTETSTRING rv_key_x;
      OCTETSTRING rv_key_y;
      OCTETSTRING rv_comp_key; // public compressed key, 33 or 49 bytes length, byte #0 indicating compressed-y-0 (0x02) or compressed-y-1 (0x03)
      security_services_its::fill_public_key_vectors(algorithm, p, rv_comp_key, rv_key_x, rv_key_y);
      // Compute the hash od the toBeSigned
      OCTETSTRING tbs;
      etsi_ts103097_tobesigned_certificate_codec codec;
      codec.encode(decoded_certificate.toBeSigned(), tbs);
      if (!tbs.is_bound()) {
        loggers::get_instance().warning("certificates_loader::load_certificate: Failed to encode toBeSigned");
        return -1;
      }
      loggers::get_instance().log_msg("certificates_loader::build_certificates_cache: Implicit encoded toBeSigned: ", tbs);
      OCTETSTRING hash_tbs;
      sha256 sha;
      sha.generate(tbs, hash_tbs);
      loggers::get_instance().log_msg("certificates_loader::build_certificates_cache: Implicit hash(toBeSignedCertificate): ", hash_tbs);
      loggers::get_instance().log_msg("certificates_loader::build_certificates_cache: Implicit issuer hash: ", m.cbegin()->second->hash_sha_256());
      OCTETSTRING input = hash_tbs + m.cbegin()->second->hash_sha_256();
      loggers::get_instance().log_msg("certificates_loader::build_certificates_cache: input: ", input);
      sha.generate(input, hash_tbs);
      loggers::get_instance().log_msg("certificates_loader::build_certificates_cache: joint hash: ", hash_tbs);
      // Reconstruction of the public key.
      OCTETSTRING public_key_x;       // public keys X-coordinate
      OCTETSTRING public_key_y;       // public keys Y-coordinate
      OCTETSTRING public_comp_key;    // public compressed key, 33 or 49 bytes length, byte #0 indicating compressed-y-0 (0x02) or compressed-y-1 (0x03)
      INTEGER public_comp_key_mode;   // public compressed key mode (02 or 03)
      security_ecc r_key(algorithm, rv_key_x, rv_key_y); // Reconstruction key
      if (r_key.reconstruct_public_keys(hash_tbs, static_cast<const OCTETSTRING&>(issuer_sign_key_x), static_cast<const OCTETSTRING&>(issuer_sign_key_y), public_key_x, public_key_y, public_comp_key, public_comp_key_mode) == -1) {
        loggers::get_instance().warning("certificates_loader::load_certificate: Failed to encode toBeSigned");
        return -1;
      }
      loggers::get_instance().log_msg("certificates_loader::load_certificate: public_key_x: ", public_key_x);
      loggers::get_instance().log_msg("certificates_loader::load_certificate: public_key_y: ", public_key_y);
      loggers::get_instance().log_msg("certificates_loader::load_certificate: public_comp_key: ", public_comp_key);
      loggers::get_instance().log_msg("certificates_loader::load_certificate: public_comp_key_mode: ", public_comp_key_mode);

      OCTETSTRING hash = hash_sha_256; // Whole-certificate hash
      OCTETSTRING hashed_id;            // Whole-certificate hashedid-8
      hashed_id = OCTETSTRING(8, static_cast<const unsigned char *>(hash) + hash.lengthof() - 8);
      loggers::get_instance().log_msg("certificates_loader::load_certificate: hash: ", hash);
      loggers::get_instance().log_msg("certificates_loader::load_certificate: hashed_id: ", hashed_id);

      // Create new record
      security_db_record *r = new security_db_record(
        algorithm,          // Signing algorithm
        key,                // Certificate name. Index key
        certificate,        // Certificate
        decoded_certificate,
        issuer,             // Hashed ID fo the issuer
        hash,               // Hash of the certificate
        hash_sha_256,       // Whole-certificate SHA256
        hashed_id,          // Hashed ID
        private_key,        // Private key
        public_key_x,       // public keys X-coordinate
        public_key_y,       // public keys Y-coordinate
        public_comp_key,    // public compressed key, 33 or 49 bytes length, byte #0 indicating compressed-y-0 (0x02) or compressed-y-1 (0x03)
        private_enc_key,    // Private encryption key
        public_enc_key_x,   // Public encryption key X-coordinate
        public_enc_key_y,   // Public encryption key Y-coordinate
        public_enc_comp_key // Public compressed encryption key
      );
      result = p_certificates.insert(std::pair<std::string, std::unique_ptr<const security_db_record>>(key, std::unique_ptr<const security_db_record>(r)));
      if (result.second == false) {
        loggers::get_instance().warning("certificates_loader::load_certificate: Failed to insert new record '%s'", key.c_str());
        delete r;
        return -1;
      }
    } else {
      loggers::get_instance().error("certificates_loader::build_certificates_cache: Unsupported verifyKeyIndicator variant");
    }
  }

  loggers::get_instance().log("<<< certificates_loader::build_certificates_cache: 0");
  return 0;
}

int certificates_loader::save_certificate(const security_db_record &p_certificate) {
  loggers::get_instance().log(">>> certificates_loader::save_certificate: certificate_id=%s", p_certificate.certificate_id().c_str());
  loggers::get_instance().log_msg("certificates_loader::save_certificate: hash=", p_certificate.hash());
  loggers::get_instance().log_msg("certificates_loader::save_certificate: hashedid8=", p_certificate.hashed_id());

  // Certificate file
  std::experimental::filesystem::path p(_full_path);
  p /= p_certificate.certificate_id();
  p += _certificateExt;
  if (std::experimental::filesystem::exists(p)) {
    std::experimental::filesystem::remove(p);
  }
  loggers::get_instance().log("certificates_loader::save_certificate: Certificate file: '%s'", p.string().c_str());
  std::ofstream os(p.string(), ios::out | ios::binary);
  os.write((char *)static_cast<const unsigned char *>(p_certificate.certificate()), p_certificate.certificate().lengthof());
  os.close();
  std::experimental::filesystem::permissions(p, std::experimental::filesystem::perms::add_perms | std::experimental::filesystem::perms::owner_all |
                                                  std::experimental::filesystem::perms::group_all | std::experimental::filesystem::perms::others_all);

  // Private key
  p = _full_path;
  p /= p_certificate.certificate_id();
  p += _privateKeyExt;
  if (std::experimental::filesystem::exists(p)) {
    std::experimental::filesystem::remove(p);
  }
  loggers::get_instance().log("certificates_loader::save_certificate: Private key file: '%s'", p.string().c_str());
  os.open(p.string(), ios::out | ios::binary);
  loggers::get_instance().log_msg("certificates_loader::save_certificate: Private key content: ", p_certificate.private_key());
  os.write((char *)static_cast<const unsigned char *>(p_certificate.private_key()), p_certificate.private_key().lengthof());
  os.close();
  std::experimental::filesystem::permissions(p, std::experimental::filesystem::perms::add_perms | std::experimental::filesystem::perms::owner_all |
                                                  std::experimental::filesystem::perms::group_all | std::experimental::filesystem::perms::others_all);

  // Private encryption key
  if (p_certificate.private_enc_key().lengthof() != 0) {
    p = _full_path;
    p /= p_certificate.certificate_id();
    p += _privateEncKeyExt;
    if (std::experimental::filesystem::exists(p)) {
      std::experimental::filesystem::remove(p);
    }
    loggers::get_instance().log("certificates_loader::save_certificate: Private encryption key file: '%s'", p.string().c_str());
    os.open(p.string(), ios::out | ios::binary);
    os.write((char *)static_cast<const unsigned char *>(p_certificate.private_enc_key()), p_certificate.private_enc_key().lengthof());
    os.close();
    std::experimental::filesystem::permissions(p, std::experimental::filesystem::perms::add_perms | std::experimental::filesystem::perms::owner_all |
                                                    std::experimental::filesystem::perms::group_all | std::experimental::filesystem::perms::others_all);
  }

  _hashed_id8s.insert(std::pair<OCTETSTRING, const std::string>(p_certificate.hashed_id(), p_certificate.certificate_id()));
  _certificates_idx.insert(std::pair<std::string, const OCTETSTRING>(p_certificate.certificate_id(), p_certificate.hashed_id()));

  return 0;
} // End of method save_certificate

